--------------------------------------------------------------------------------
-- AVALON SCALER
--------------------------------------------------------------------------------
-- DO 10/2018
--------------------------------------------------------------------------------
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.ALL;

USE std.textio.ALL;

LIBRARY work;
USE work.base_pack.ALL;

ENTITY tb_ascal_test IS

END ENTITY tb_ascal_test;

--------------------------------------------------------------------------------

ARCHITECTURE beh OF tb_ascal_test IS
  
  CONSTANT FRAC  : natural :=4;

--  CONSTANT N_DW  : natural :=64;
  CONSTANT N_DW  : natural :=128;
  
  COMPONENT ascal IS
    GENERIC (
    MASK    : unsigned(0 TO 7) :=x"FF";
    RAMBASE   : unsigned(31 DOWNTO 0) := x"0000_0000";
    RAMSIZE   : unsigned(31 DOWNTO 0) := x"0080_0000"; -- =8MB
    INTER   : boolean := true;
    HEADER   : boolean := false;
    BYTESWAP : boolean := false;
    FRAC    : natural RANGE 4 TO 6 :=4;
    OHRES   : natural RANGE 1 TO 4096 :=4096;
    N_DW    : natural RANGE 32 TO 128;
    N_AW    : natural RANGE 8 TO 32;
    N_BURST : natural := 256 -- 256 bytes per burst
    );
  PORT (
    i_r               : IN  unsigned(7 DOWNTO 0);
    i_g               : IN  unsigned(7 DOWNTO 0);
    i_b               : IN  unsigned(7 DOWNTO 0);
--    i_pe              : IN  std_logic;
    i_hs              : IN  std_logic;
    i_vs              : IN  std_logic;
    i_fl              : IN  std_logic;
    i_de              : IN  std_logic;
    i_ce              : IN  std_logic;
    i_clk             : IN  std_logic;
    o_r               : OUT unsigned(7 DOWNTO 0);
    o_g               : OUT unsigned(7 DOWNTO 0);
    o_b               : OUT unsigned(7 DOWNTO 0);
    o_hs              : OUT std_logic;
    o_vs              : OUT std_logic;
    o_de              : OUT std_logic;
    o_ce              : IN  std_logic;
    o_clk             : IN  std_logic;
    iauto             : IN  std_logic;
    himin             : IN  natural RANGE 0 TO 4095;
    himax             : IN  natural RANGE 0 TO 4095;
    vimin             : IN  natural RANGE 0 TO 4095;
    vimax             : IN  natural RANGE 0 TO 4095;
    run               : IN  std_logic;
    freeze            : IN  std_logic;
    mode              : IN  unsigned(4 DOWNTO 0);
    format            : IN  unsigned(1 DOWNTO 0);
    htotal            : IN  natural RANGE 0 TO 4095;
    hsstart           : IN  natural RANGE 0 TO 4095;
    hsend             : IN  natural RANGE 0 TO 4095;
    hdisp             : IN  natural RANGE 0 TO 4095;
    hmin              : IN  natural RANGE 0 TO 4095;
    hmax              : IN  natural RANGE 0 TO 4095;
    vtotal            : IN  natural RANGE 0 TO 4095;
    vsstart           : IN  natural RANGE 0 TO 4095;
    vsend             : IN  natural RANGE 0 TO 4095;
    vdisp             : IN  natural RANGE 0 TO 4095;
    vmin              : IN  natural RANGE 0 TO 4095;
    vmax              : IN  natural RANGE 0 TO 4095;
    poly_clk          : IN std_logic;
    poly_dw           : IN unsigned(8 DOWNTO 0);
    poly_a            : IN unsigned(FRAC+2 DOWNTO 0);
    poly_wr           : IN std_logic;
    avl_clk           : IN  std_logic;
    avl_waitrequest   : IN  std_logic;
    avl_readdata      : IN  std_logic_vector(N_DW-1 DOWNTO 0);
    avl_readdatavalid : IN  std_logic;
    avl_burstcount    : OUT std_logic_vector(7 DOWNTO 0);
    avl_writedata     : OUT std_logic_vector(N_DW-1 DOWNTO 0);
    avl_address       : OUT std_logic_vector(N_AW-1 DOWNTO 0);
    avl_write         : OUT std_logic;
    avl_read          : OUT std_logic;
    avl_byteenable    : OUT std_logic_vector(N_DW/8-1 DOWNTO 0);
    reset_na          : IN  std_logic);
  END COMPONENT ascal;

  COMPONENT emu IS
    PORT (
      clk_50m          : IN    std_logic;
      reset            : IN    std_logic;
      hps_bus          : INOUT std_logic_vector(44 DOWNTO 0);
      sconf            : OUT   std_logic_vector(4 DOWNTO 0);
      clk_video        : OUT   std_logic;
      ce_pixel         : OUT   std_logic;
      video_arx        : OUT   std_logic_vector(7 DOWNTO 0);
      video_ary        : OUT   std_logic_vector(7 DOWNTO 0);
      vga_r            : OUT   std_logic_vector(7 DOWNTO 0);
      vga_g            : OUT   std_logic_vector(7 DOWNTO 0);
      vga_b            : OUT   std_logic_vector(7 DOWNTO 0);
      vga_hs           : OUT   std_logic;
      vga_vs           : OUT   std_logic;
      vga_fl           : OUT   std_logic;
      vga_de           : OUT   std_logic;
      led_user         : OUT   std_logic;
      led_power        : OUT   std_logic_vector(1 DOWNTO 0);
      led_disk         : OUT   std_logic_vector(1 DOWNTO 0);
      audio_l          : OUT   std_logic_vector(15 DOWNTO 0);
      audio_r          : OUT   std_logic_vector(15 DOWNTO 0);
      audio_s          : OUT   std_logic;
      audio_mix        : OUT   std_logic_vector(1 DOWNTO 0);
      tape_in          : IN    std_logic;
      sd_sck           : OUT   std_logic := 'Z';
      sd_mosi          : OUT   std_logic := 'Z';
      sd_miso          : IN    std_logic;
      sd_cs            : OUT   std_logic := 'Z';
      sd_cd            : IN    std_logic;
      iauto            : OUT   std_logic;
      himin            : OUT   natural RANGE 0 TO 4095;
      himax            : OUT   natural RANGE 0 TO 4095;
      vimin            : OUT   natural RANGE 0 TO 4095;
      vimax            : OUT   natural RANGE 0 TO 4095;
      hdisp            : IN    natural RANGE 0 TO 4095;
      hmin             : OUT   natural RANGE 0 TO 4095;
      hmax             : OUT   natural RANGE 0 TO 4095;
      vdisp            : IN    natural RANGE 0 TO 4095;
      vmin             : OUT   natural RANGE 0 TO 4095;
      vmax             : OUT   natural RANGE 0 TO 4095;
      ddram_clk        : OUT   std_logic;
      ddram_busy       : IN    std_logic;
      ddram_burstcnt   : OUT   std_logic_vector(7 DOWNTO 0);
      ddram_addr       : OUT   std_logic_vector(28 DOWNTO 0);
      ddram_dout       : IN    std_logic_vector(63 DOWNTO 0);
      ddram_dout_ready : IN    std_logic;
      ddram_rd         : OUT   std_logic;
      ddram_din        : OUT   std_logic_vector(63 DOWNTO 0);
      ddram_be         : OUT   std_logic_vector(7 DOWNTO 0);
      ddram_we         : OUT   std_logic;
      sdram_clk        : OUT   std_logic;
      sdram_cke        : OUT   std_logic;
      sdram_a          : OUT   std_logic_vector(12 DOWNTO 0);
      sdram_ba         : OUT   std_logic_vector(1 DOWNTO 0);
      sdram_dq         : INOUT std_logic_vector(15 DOWNTO 0);
      sdram_dqml       : OUT   std_logic;
      sdram_dqmh       : OUT   std_logic;
      sdram_ncs        : OUT   std_logic;
      sdram_ncas       : OUT   std_logic;
      sdram_nras       : OUT   std_logic;
      sdram_nwe        : OUT   std_logic);
  END COMPONENT emu;

  SIGNAL i_r             : std_logic_vector(7 DOWNTO 0);
  SIGNAL i_g             : std_logic_vector(7 DOWNTO 0);
  SIGNAL i_b             : std_logic_vector(7 DOWNTO 0);
  SIGNAL i_pe            : std_logic;
  SIGNAL i_hs            : std_logic;
  SIGNAL i_vs            : std_logic;
  SIGNAL i_fl            : std_logic :='0';
  SIGNAL i_de            : std_logic;
  SIGNAL i_ce            : std_logic;
  SIGNAL i_hblank,i_vblank : std_logic;
  SIGNAL i_clk           : std_logic :='0';
  SIGNAL o_r             : unsigned(7 DOWNTO 0);
  SIGNAL o_g             : unsigned(7 DOWNTO 0);
  SIGNAL o_b             : unsigned(7 DOWNTO 0);
  SIGNAL o_hs            : std_logic;
  SIGNAL o_vs            : std_logic;
  SIGNAL o_de            : std_logic;
  SIGNAL o_ce            : std_logic;
  SIGNAL o_clk           : std_logic :='0';
  SIGNAL run             : std_logic;
  SIGNAL freeze          : std_logic :='0';
  SIGNAL mode            : unsigned(4 DOWNTO 0);
  SIGNAL format          : unsigned(1 DOWNTO 0);
  SIGNAL htotal          : natural RANGE 0 TO 4095;
  SIGNAL hsstart         : natural RANGE 0 TO 4095;
  SIGNAL hsend           : natural RANGE 0 TO 4095;
  SIGNAL hdisp           : natural RANGE 0 TO 4095;
  SIGNAL hmin            : natural RANGE 0 TO 4095;
  SIGNAL hmax            : natural RANGE 0 TO 4095;
  SIGNAL vtotal          : natural RANGE 0 TO 4095;
  SIGNAL vsstart         : natural RANGE 0 TO 4095;
  SIGNAL vsend           : natural RANGE 0 TO 4095;
  SIGNAL vdisp           : natural RANGE 0 TO 4095;
  SIGNAL vmin            : natural RANGE 0 TO 4095;
  SIGNAL vmax            : natural RANGE 0 TO 4095;
  SIGNAL avl_clk         : std_logic :='0';
  SIGNAL avl_waitrequest : std_logic;
  SIGNAL avl_readdata    : std_logic_vector(N_DW-1 DOWNTO 0);
  SIGNAL avl_readdatavalid : std_logic;
  SIGNAL avl_burstcount  : std_logic_vector(7 DOWNTO 0);
  SIGNAL avl_writedata   : std_logic_vector(N_DW-1 DOWNTO 0);
  SIGNAL avl_address     : std_logic_vector(27 DOWNTO 0);
  SIGNAL avl_write       : std_logic;
  SIGNAL avl_read        : std_logic;
  SIGNAL avl_byteenable  : std_logic_vector(N_DW/8-1 DOWNTO 0);
  SIGNAL reset_na        : std_logic;

  SIGNAL poly_clk     : std_logic;
  SIGNAL poly_dw      : unsigned(8 DOWNTO 0);
  SIGNAL poly_a       : unsigned(FRAC+2 DOWNTO 0);
  SIGNAL poly_wr      : std_logic;

  FILE fil : text OPEN write_mode IS "mem.log";
  FILE fvic : text OPEN write_mode IS "videi.log";
  FILE fvir : text OPEN write_mode IS "vider.log";
  FILE fvid : text OPEN write_mode IS "video.log";
  FILE fvif : text OPEN write_mode IS "videm.log";

  CONSTANT uu : unsigned(N_DW-1 DOWNTO 0) :=(OTHERS =>'1');
  TYPE arr_uv_dw IS ARRAY (natural RANGE <>) OF unsigned(N_DW-1 DOWNTO 0);
  SHARED VARIABLE mem : arr_uv_dw(0 TO 2**17-1) := (OTHERS =>uu);
  
  SIGNAL iauto   : std_logic;
  SIGNAL himin   : natural RANGE 0 TO 4095;
  SIGNAL himax   : natural RANGE 0 TO 4095;
  SIGNAL vimin   : natural RANGE 0 TO 4095;
  SIGNAL vimax   : natural RANGE 0 TO 4095;
  
  SIGNAL clk_50m : std_logic :='0';
  SIGNAL clk_25m : std_logic :='0';
  SIGNAL clk_12m5 : std_logic :='0';
  SIGNAL i_shift : std_logic_vector(3 DOWNTO 0) :="0001";
  SIGNAL reset : std_logic :='0';
  SIGNAL hps_bus : std_logic_vector(44 DOWNTO 0);
  SIGNAL sconf : std_logic_vector(4 DOWNTO 0);
  
  COMPONENT osd
    PORT (
      clk_sys : IN  std_logic;
      io_osd  : IN  std_logic;
      io_strobe : IN std_logic;
      io_din : IN unsigned(15 DOWNTO 0);
      clk_video : IN std_logic;
      din     : IN unsigned(23 DOWNTO 0);
      dout    : OUT unsigned(23 DOWNTO 0);
      de_in   : IN std_logic;
      de_out  : OUT std_logic);
  END COMPONENT osd;

  COMPONENT screen_rotate
    GENERIC (
      WIDTH    : natural := 256;
      HEIGHT   : natural := 224;
      DEPTH    : natural := 8;
      MARGIN   : natural := 8;
      CCW      : natural := 0;
      DOUBLING : natural := 0);
    PORT (
      clk_in    : IN  std_logic;
      ce_in     : IN  std_logic;
      video_in  : IN  unsigned(DEPTH-1 DOWNTO 0);
      hblank    : IN  std_logic;
      vblank    : IN  std_logic;
      clk_out   : IN  std_logic;
      video_out : OUT unsigned(DEPTH-1 DOWNTO 0);
      hsync     : OUT std_logic;
      vsync     : OUT std_logic;
      de        : OUT std_logic);
  END COMPONENT;
  
  SIGNAL oo_de : std_logic;
  SIGNAL din,dout : unsigned(23 DOWNTO 0);
  SIGNAL r_r : unsigned(7 DOWNTO 0);
  SIGNAL r_hs,r_vs,r_de : std_logic;
  SIGNAL ri_hb,ri_vb : std_logic;
  SIGNAL i_vblank1 : std_logic;
  SIGNAL i_hblank1,i_hblank2 : std_logic;
  SIGNAL r_de1,r_de2,r_hs1,r_hs2,r_vs1,r_vs2 : std_logic;
  SIGNAL rde_re,rde_delay : time;
BEGIN

  avl_clk<=NOT avl_clk AFTER 0.5 us / 80; --80;
  clk_50m<=NOT clk_50m AFTER 0.5 us / 120;--50;

  i_pe<=i_de; -- WHEN rising_edge(clk_12m5);

  reset_na<='0','1' AFTER 1.1 us;

--screen_rotate #(256,224,8) screen_rotate
  rde_re<=now WHEN rising_edge(r_de);
  rde_delay<=now-rde_re WHEN falling_edge(r_de);
  
  ----------------------------------------------------------
  --i_rotate: screen_rotate
  --  GENERIC MAP (
  --    WIDTH    =>256,
  --    HEIGHT   =>224,
  --    DEPTH    =>8)
  --  PORT MAP (
  --    clk_in    => clk_50m,
  --    ce_in     => i_ce,
  --    video_in  => unsigned(i_r),
  --    hblank    => i_hblank,
  --    vblank    => i_vblank,
  --    clk_out   => clk_50m,
  --    video_out => r_r,
  --    hsync     => r_hs,
  --    vsync     => r_vs,
  --    de        => r_de);

  --PROCESS(clk_50m) IS
  --BEGIN
  --  IF rising_edge(clk_50m) THEN
  --    IF i_ce='1' THEN
  --      i_vblank1<=i_vblank;
  --      i_hblank1<=i_hblank;
  --    END IF;
  --    r_de1<=r_de;
  --    r_hs1<=r_hs;
  --    r_vs1<=r_vs;
  --    r_de2<=r_de1;
  --    r_hs2<=r_hs1;
  --    r_vs2<=r_vs1;
      
  --  END if;
  --END process;
  
  ----------------------------------------------------------
  i_ascal: ascal
    GENERIC MAP (
      N_DW    => N_DW,
      N_AW    => 28,
      N_BURST => 256)
    PORT MAP (
      
      i_r               => unsigned(i_r),
      i_g               => unsigned(i_g),
      i_b               => unsigned(i_b),
     -- i_pe              => i_pe,
      i_hs              => i_hs,
      i_vs              => i_vs,
      i_fl              => i_fl,
      i_de              => i_de,
      --i_r => r_r,
      --i_g => r_r,
      --i_b => r_r,
      --i_hs => r_hs2,
      --i_vs => r_vs2,
      --i_fl => '1',
      --i_de => r_de2,
      i_ce              => i_ce,
      i_clk             => clk_12m5, --i_clk, --clk_50m, --i_clk,
      o_r               => o_r,
      o_g               => o_g,
      o_b               => o_b,
      o_hs              => o_hs,
      o_vs              => o_vs,
      o_de              => o_de,
      o_ce              => o_ce,
      o_clk             => o_clk,
      iauto             => '1', --iauto,
      himin             => himin,
      himax             => himax,
      vimin             => vimin,
      vimax             => vimax,
      run               => run,
      freeze            => freeze,
      mode              => mode,
      format            => format,
      htotal            => htotal,
      hsstart           => hsstart,
      hsend             => hsend,
      hdisp             => hdisp,
      hmin              => hmin,
      hmax              => hmax,
      vtotal            => vtotal,
      vsstart           => vsstart,
      vsend             => vsend,
      vdisp             => vdisp,
      vmin              => vmin,
      vmax              => vmax,
      poly_clk          => poly_clk,
      poly_dw           => poly_dw,
      poly_a            => poly_a,
      poly_wr           => poly_wr,
      avl_clk           => avl_clk,
      avl_waitrequest   => avl_waitrequest,
      avl_readdata      => avl_readdata,
      avl_readdatavalid => avl_readdatavalid,
      avl_burstcount    => avl_burstcount,
      avl_writedata     => avl_writedata,
      avl_address       => avl_address,
      avl_write         => avl_write,
      avl_read          => avl_read,
      avl_byteenable    => avl_byteenable,
      reset_na          => reset_na);
  
  clk_25m<=NOT clk_25m WHEN rising_edge(clk_50m);
  clk_12m5<=NOT clk_12m5 WHEN rising_edge(clk_25m);

  --i_ce<=i_shift(0);
  --i_shift<=i_shift(2 DOWNTO 0) & i_shift(3) WHEN rising_edge(clk_50m);
  i_emu: emu
    PORT MAP (
      clk_50m          => clk_12m5,
      reset            => reset,
      hps_bus          => hps_bus,
      sconf            => sconf,
      clk_video        => i_clk,
      ce_pixel         => i_ce,
      video_arx        => OPEN,
      video_ary        => OPEN,
      vga_r            => i_r,
      vga_g            => i_g,
      vga_b            => i_b,
      vga_hs           => i_hs,
      vga_vs           => i_vs,
      vga_fl           => i_fl,
      vga_de           => i_de,
      led_user         => OPEN,
      led_power        => OPEN, 
      led_disk         => OPEN,
      audio_l          => OPEN,
      audio_r          => OPEN ,
      audio_s          => OPEN,
      audio_mix        => OPEN,
      tape_in          => '0',
      sd_sck           => OPEN,
      sd_mosi          => OPEN,
      sd_miso          => '0',
      sd_cs            => OPEN,
      sd_cd            => '0',
      iauto            => iauto,
      himin            => himin,
      himax            => himax,
      vimin            => vimin,
      vimax            => vimax,
      hdisp            => hdisp,
      hmin             => hmin,
      hmax             => hmax,
      vdisp            => vdisp,
      vmin             => vmin,
      vmax             => vmax,
      ddram_clk        => OPEN,
      ddram_busy       => '0',
      ddram_burstcnt   => OPEN,
      ddram_addr       => OPEN,
      ddram_dout       => x"00000000_00000000",
      ddram_dout_ready => '0',
      ddram_rd         => OPEN,
      ddram_din        => OPEN,
      ddram_be         => OPEN,
      ddram_we         => OPEN,
      sdram_clk        => OPEN,
      sdram_cke        => OPEN,
      sdram_a          => OPEN,
      sdram_ba         => OPEN,
      sdram_dq         => OPEN,
      sdram_dqml       => OPEN,
      sdram_dqmh       => OPEN,
      sdram_ncs        => OPEN,
      sdram_ncas       => OPEN,
      sdram_nras       => OPEN,
      sdram_nwe        => OPEN);
  
  --i_osd : osd
  --  PORT MAP (
  --  clk_sys => clk_50m,
  --  io_osd => '0',
  --  io_strobe => '0',
  --  io_din => x"0000",
  --  clk_video => o_clk,
  --  din => din,
  --  dout => dout,
  --  de_in => o_de,
  --  de_out => oo_de);

  --din<=o_r & o_g & o_b;
  
  
  ----------------------------------------------------------
  Poly:PROCESS IS
    TYPE arr_integer IS ARRAY (natural RANGE <>) OF integer;
    CONSTANT TABLE_SCANLINE : arr_integer :=(
      -- This is Catmull-Rom with 80% Scanlines
      0, 128,   0,   0,
      -4, 127,   5,  -0,
      -6, 123,  12,  -1,
      -8, 118,  20,  -2,
      -9, 111,  29,  -3,
      -9, 102,  39,  -4,
      -9,  93,  50,  -6,
      -9,  83,  61,  -7,
      -8,  72,  72,  -8,
      -7,  61,  83,  -9,
      -6,  50,  93,  -9,
      -4,  39, 102,  -9,
      -3,  29, 111,  -9,
      -2,  20, 118,  -8,
      -1,  12, 123,  -6,
      -0,   5, 127,  -4,

      0, 128,   0,   0,
      0, 124,   0,   0,
      0, 113,   0,   0,
      0,  96,   1,   0,
      0,  75,   1,   0,
      0,  54,   3,   0,
      0,  36,   5,   0,
      0,  22,   8,   0,
      0,  13,  13,   0,
      0,   8,  22,   0,
      0,   5,  36,   0,
      0,   3,  54,   0,
      0,   1,  75,   0,
      0,   1,  96,   0,
      0,   0, 113,   0,
      0, 0, 124, 0);

    CONSTANT TABLE_NEAREST : arr_integer :=(
      0, 128,   0,   0,
      0, 128,   0,   0,
      0, 128,   0,   0,
      0, 128,   0,   0,
      0, 128,   0,   0,
      0, 128,   0,   0,
      0, 128,   0,   0,
      0, 128,   0,   0,
      0,   0, 128,   0,
      0,   0, 128,   0,
      0,   0, 128,   0,
      0,   0, 128,   0,
      0,   0, 128,   0,
      0,   0, 128,   0,
      0,   0, 128,   0,
      0,   0, 128,   0,
      0, 128,   0,   0,
      0, 128,   0,   0,
      0, 128,   0,   0,
      0, 128,   0,   0,
      0, 128,   0,   0,
      0, 128,   0,   0,
      0, 128,   0,   0,
      0, 128,   0,   0,
      0,   0, 128,   0,
      0,   0, 128,   0,
      0,   0, 128,   0,
      0,   0, 128,   0,
      0,   0, 128,   0,
      0,   0, 128,   0,
      0,   0, 128,   0,
      0, 0, 128, 0);
    
  BEGIN
    FOR i IN 0 TO 16*5-1 LOOP
      poly_a<=to_unsigned(i,7);
      poly_dw<=unsigned(to_signed(TABLE_SCANLINE(i),9));
--      poly_dw<=unsigned(to_signed(TABLE_NEAREST(i),9));
      poly_wr<='1';
      wure(avl_clk);
    END LOOP;
    poly_wr<='0';
    wure(avl_clk);
    WAIT;
    
  END PROCESS;
  poly_clk<=avl_clk;
  ----------------------------------------------------------
  AvaleW:PROCESS IS
    VARIABLE a : unsigned(16 DOWNTO 0);
  BEGIN
    wure(avl_clk);
    IF avl_write='1' AND avl_waitrequest='0' THEN
      a:=unsigned(avl_address(16 DOWNTO 0));
      FOR i IN 0 TO to_integer(unsigned(avl_burstcount))-1 LOOP
        mem(to_integer(a)+i):=unsigned(avl_writedata);
        wure(avl_clk);
      END LOOP;
    END IF;
  END PROCESS AvaleW;

  AvaleR:PROCESS IS
--    VARIABLE a : unsigned(15 DOWNTO 0);
    VARIABLE lev : natural RANGE 0 TO 2 :=0;
    VARIABLE am,am2 : unsigned(16 DOWNTO 0);
    VARIABLE i : natural RANGE 0 TO 255 :=0;
    VARIABLE delay : natural RANGE 0 TO 250;
    VARIABLE lm,lm2 : natural RANGE 0 TO 255;
  BEGIN
    wure(avl_clk);
    avl_readdatavalid<='0';

    IF avl_read='1' AND avl_waitrequest='0' THEN
      IF lev=0 THEN
        am:=unsigned(avl_address(16 DOWNTO 0));
        lm:=to_integer(unsigned(avl_burstcount));
        lev:=1;
      ELSIF lev=1 THEN
        am2:=unsigned(avl_address(16 DOWNTO 0));
        lm2:=to_integer(unsigned(avl_burstcount));
        lev:=2;
      END IF;
    END IF;

    IF lev>0 THEN
      delay:=delay+1;
    END IF;
    IF lev>0 AND delay>20 THEN
      avl_readdata<=std_logic_vector(mem(to_integer(am)+i));
      avl_readdatavalid<='1';
      i:=i+1;
      IF i=lm THEN
        lev:=lev-1;
        am:=am2;
        lm:=lm2;
        i:=0;
        delay:=0;
      END IF;
    END IF;
    
    --IF avl_read='1' AND avl_waitrequest='0' THEN
    --  a:=unsigned(avl_address(15 DOWNTO 0));
    --  wure(avl_clk,40);
    --  FOR i IN 0 TO 7 LOOP
    --    avl_readdata<=std_logic_vector(mem(to_integer(a)+i));
    --    avl_readdatavalid<='1';
    --    wure(avl_clk);
    --  END LOOP;
    --  --avl_readdatavalid<='0';
    --  --wure(avl_clk);
    --  --avl_readdatavalid<='0';
    --  --wure(avl_clk);
    --  --avl_readdatavalid<='0';
    --  --wure(avl_clk);
      
    --  FOR i IN 8 TO to_integer(unsigned(avl_burstcount))-1 LOOP
    --    avl_readdata<=std_logic_vector(mem(to_integer(a)+i));
    --    avl_readdatavalid<='1';
    --    wure(avl_clk);
    --  END LOOP;
      
    --END IF;
  END PROCESS AvaleR;
  
  PROCESS IS
  BEGIN
    wure(avl_clk);
    avl_waitrequest<='0';
    WAIT;
    --wure(avl_clk,5);
    --avl_waitrequest<='0';
    --wure(avl_clk,2);
    --avl_waitrequest<='1';
    --wure(avl_clk,3);
    --avl_waitrequest<='0';
  END PROCESS;
  
  ----------------------------------------------------------
  --Aargh:IF 0=1 GENERATE
  Trace:PROCESS IS
    VARIABLE rad,wad : unsigned(27 DOWNTO 0);
    VARIABLE rcpt,rlen,wcpt,wlen : natural RANGE 0 TO 256-1;
    VARIABLE rdata,wdata : arr_uv_dw(0 TO 16*4-1);
    VARIABLE tdata : arr_uv8(0 TO 65535);
    VARIABLE rpend,wpend : boolean;
    VARIABLE lout : line;
    VARIABLE k : natural := 0;
    VARIABLE KLEN : natural := 16*16*4;
  BEGIN
    wure(avl_clk);
    IF avl_read='1' AND avl_waitrequest='0' THEN
      rad:=unsigned(avl_address);
      rpend:=true;
      rcpt:=0;
      rlen:=to_integer(unsigned(avl_burstcount));
    END IF;
    IF rpend AND avl_readdatavalid='1' THEN
      rdata(rcpt):=unsigned(avl_readdata);
      rcpt:=rcpt+1;
      IF rcpt=rlen THEN
        rpend:=false;
        write(lout,string'("R : " & to_hstring(rad) & ": "));
        FOR i IN 0 TO rlen-1 LOOP
          write(lout,to_hstring(rdata(i)) & " ");
        END LOOP;
        write (lout,now);
        writeline(fil,lout);
      END IF;
    END IF;
    IF avl_write='1' AND avl_waitrequest='0' THEN
      IF NOT wpend THEN
        wad:=unsigned(avl_address);
        wpend:=true;
        wcpt:=1;
        wlen:=to_integer(unsigned(avl_burstcount));
        wdata(0):=unsigned(avl_writedata);
      ELSE
        wdata(wcpt):=unsigned(avl_writedata);
        wcpt:=wcpt+1;
        IF wcpt=wlen THEN
          IF wad=0 OR wad=16 OR k>=KLEN THEN
            FOR i IN 0 TO k/3 LOOP
              IF (tdata((i*3) MOD 65536))=x"FF" THEN
                write (lout,string'("#"));
              ELSE
                write (lout,string'(" "));
              END IF;
              --write(lout,to_hstring(tdata((i*6) MOD 65536)));
            END LOOP;
            writeline(fvif,lout);
            k:=0;
          END IF;
          write(lout,string'("W : " & to_hstring(wad) & ": "));
          FOR i IN 0 TO wlen-1 LOOP
            write(lout,to_hstring(wdata(i)) & " ");
            FOR j IN 0 TO N_DW/8-1 LOOP
              tdata(k MOD 65536):=unsigned(wdata(i)(N_DW-1-j*8 DOWNTO N_DW-8-j*8));
              k:=k+1;
            END LOOP;
          END LOOP;
          write (lout,now);
          writeline(fil,lout);
          wpend:=false;
        END IF;
      END IF;
    END IF;
    
  END PROCESS Trace;
  --END GENERATE;
  ----------------------------------------------------------
  --Ooops:IF 0=1 GENERATE
  VideoI:PROCESS IS
    VARIABLE pix : string(1 TO 10000) :=(OTHERS =>nul);
    VARIABLE i : natural :=1;
    VARIABLE rgb : unsigned(2 DOWNTO 0);
    VARIABLE lout : line;
    VARIABLE l : natural;
  BEGIN
    WAIT UNTIL rising_edge(i_clk);
    WAIT FOR 0.1 ns;
    IF i_ce='1' THEN
      rgb:=i_r(7) & i_g(7) & i_b(7);
      IF i_de='1' THEN
        CASE rgb IS
          WHEN "000"  => pix(i):=' ';
          WHEN "001"  => pix(i):='B';
          WHEN "010"  => pix(i):='G';
          WHEN "011"  => pix(i):='C';
          WHEN "100"  => pix(i):='R';
          WHEN "101"  => pix(i):='M';
          WHEN "110"  => pix(i):='Y';
          WHEN "111"  => pix(i):='W';
          WHEN OTHERS => pix(i):='?';
        END CASE;
      ELSE
        pix(i):='.';
      END IF;
      i:=i+1;
      
      IF i_hs='1' THEN
        write (lout,pix(1 TO i-1));
        write (lout,string'(" "));
        write (lout,l);
        write (lout,string'(" "));
        write (lout,now);
        writeline(fvic,lout);
        WAIT UNTIL i_hs='0';
        i:=1;
        l:=l+1;
        pix:=(OTHERS =>nul);
      END IF;

      IF i_vs='1' THEN
        i:=1;
        l:=0;
        WAIT UNTIL i_vs='0';
      END IF;
    END IF;
  END PROCESS;
  --END GENERATE;
  
  ----------------------------------------------------------
  --Ooops:IF 0=1 GENERATE
  VideoR:PROCESS IS
    VARIABLE pix : string(1 TO 10000) :=(OTHERS =>nul);
    VARIABLE i : natural :=1;
    VARIABLE rgb : unsigned(2 DOWNTO 0);
    VARIABLE lout : line;
    VARIABLE l : natural;
  BEGIN
    WAIT UNTIL rising_edge(clk_50m);
    WAIT FOR 0.1 ns;
    IF true THEN
      rgb:=r_r(7) & r_r(7) & r_r(7);
      IF r_de='1' THEN
        CASE rgb IS
          WHEN "000"  => pix(i):=' ';
          WHEN "001"  => pix(i):='B';
          WHEN "010"  => pix(i):='G';
          WHEN "011"  => pix(i):='C';
          WHEN "100"  => pix(i):='R';
          WHEN "101"  => pix(i):='M';
          WHEN "110"  => pix(i):='Y';
          WHEN "111"  => pix(i):='W';
          WHEN OTHERS => pix(i):='?';
        END CASE;
      ELSE
        pix(i):='.';
      END IF;
      i:=i+1; IF i>10000 THEN i:=10000; END IF;
      
      IF r_hs='1' THEN
        write (lout,pix(1 TO i-1));
        write (lout,string'(" "));
        write (lout,l);
        write (lout,string'(" "));
        write (lout,now);
        writeline(fvir,lout);
        WAIT UNTIL r_hs='0';
        i:=1;
        l:=l+1;
        pix:=(OTHERS =>nul);
      END IF;

      IF r_vs='1' THEN
        i:=1;
        l:=0;
        WAIT UNTIL r_vs='0';
      END IF;
    END IF;
  END PROCESS;
  --END GENERATE;
  
  ----------------------------------------------------------
  VideoO:PROCESS IS
    VARIABLE pix : string(1 TO 10000) :=(OTHERS =>nul);
    VARIABLE i : natural :=1;
    VARIABLE rgb : unsigned(2 DOWNTO 0);
    VARIABLE lout : line;
    VARIABLE l : natural;
  BEGIN
    WAIT UNTIL rising_edge(o_clk);
    WAIT FOR 0.1 ns;
    IF o_ce='1' THEN
      rgb:=o_r(7) & o_g(7) & o_b(7);
      IF o_de='1' THEN
        IF o_r/=o_b OR o_r/=o_g OR o_g/=o_b THEN
          CASE rgb IS
            WHEN "000"  => pix(i):=' ';
            WHEN "001"  => pix(i):='B';
            WHEN "010"  => pix(i):='G';
            WHEN "011"  => pix(i):='C';
            WHEN "100"  => pix(i):='R';
            WHEN "101"  => pix(i):='M';
            WHEN "110"  => pix(i):='Y';
            WHEN "111"  => pix(i):='W';
                           IF o_r(7 DOWNTO 6)="11" THEN Pix(i):='#';
                           ELSE pix(i):='+';
                           END IF;
            WHEN OTHERS => pix(i):='?';
          END CASE;
        ELSE
          CASE o_r(7 DOWNTO 5) IS
            WHEN "111" => pix(i):='#';
            WHEN "110" => pix(i):='X';
            WHEN "101" => pix(i):='*';
            WHEN "100" => pix(i):='+';
            WHEN "011" => pix(i):='-';
            WHEN "010" => pix(i):='.';
            WHEN "001" => pix(i):=' ';
            WHEN OTHERS => pix(i):=' ';
          END CASE;
        END IF;
      ELSE
        pix(i):='.';
      END IF;
      i:=i+1;
      
      IF o_hs='1' THEN
        write (lout,pix(1 TO i-1));
        write (lout,string'(" "));
        write (lout,l);
        write (lout,string'(" "));
        write (lout,now);
        writeline(fvid,lout);
        WAIT UNTIL o_hs='0';
        i:=1;
        l:=l+1;
        pix:=(OTHERS =>nul);
      END IF;

      IF o_vs='1' THEN
        i:=1;
        l:=0;
        WAIT UNTIL o_vs='0';
      END IF;
    END IF;
  END PROCESS;
  
--  ----------------------------------------------------------
 
  o_clk<=NOT o_clk AFTER 0.5 us / 65;
  o_ce<='1';

--  --1920x1080@60 PCLK=148.5MHz CEA
----reg  [11:0] WIDTH  = 1920;
----reg  [11:0] HFP    = 88;
----reg  [11:0] HS     = 48;
----reg  [11:0] HBP    = 148;
----reg  [11:0] HEIGHT = 1080;
----reg  [11:0] VFP    = 4;
----reg  [11:0] VS     = 5;
----reg  [11:0] VBP    = 36;
----reg  [11:0] VSET   = 0;
--  --hdisp<=1920; --1920; --602+4; --*4;  --800; --320; --1000; --1920;
--  --vdisp<=1200; --1080; --256; --800; --600; --600; --800; -- 768; -- 1080;
--  -- 155 < rien
--  -- 156 < rien
--  -- 157 < rien
--  -- 158 < rien
--  -- 159 < rien
--  -- 160 < rien
--  -- 161 < rien
--  -- 162 < rien
--  -- 163 < rien
  hdisp<=1280; --512; --07; --200; --1920; --800; --1920*2; --640;
  vdisp<=896;  --300;
  
--For which core?  For Genesis with scandoublerfx on the input to the scaler is 640x448. The display resolution is 1920x1080 but with vscale_mode=1 mister should be producing an image 1280x896 letterboxed into the 1920x1080.  The width can vary based on aspect ratio but horizontal scaling isn't where we notice the problem.
  
  --hsstart<=hdisp+88;
  --hsend<=hdisp+88+48;
  --htotal<=hdisp+88+48+48;
  
  vsstart<=vdisp+4;
  vsend<=vdisp+4+5;
  vtotal<=vdisp+4+5+36+10;


  hsstart<=hdisp+4;
  hsend<=hdisp+8;
  htotal<=hdisp+16+10+20+100;
  
  
  
  run<='1';

  mode<="00100";
  format<="10";
  
END ARCHITECTURE beh;
